/*
 * Copyright (C) IBM Corp. 2005
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation; either version 2.1 of the
 *  License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

/*
 * scsi_blacklist:
 *
 * By default, reads black/white list data from all files under /etc/scsi/
 * that end in ".devinfo", and writes the output to /proc/scsi/device_info.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <dirent.h>
#include <errno.h>

#include "scsi_devinfo.h"

#define	DEVINFO_OUT	"/proc/scsi/device_info"
#define	SCSI_DIR	"/etc/scsi"
#define DEVINFO_SUFFIX		".devinfo"

#define	MAX_NAME_LEN	128
#define GETLINE_CHUNK_SIZE	128

static const char options[] = "d:p:tV";

static char scsidir[MAX_NAME_LEN] = SCSI_DIR;
static char out_file[MAX_NAME_LEN] = DEVINFO_OUT;
static int test_only; /* defaults to zero */
static int debug; /* defaults to zero */

static char cur_filename[MAX_NAME_LEN];
static int cur_lineno;

#define ENTRY(value)	#value, value
const struct {
	char *name;
	int value;
} devinfo_to_value[] = {
	{ ENTRY(BLIST_NOLUN) },
	{ ENTRY(BLIST_FORCELUN) },
	{ ENTRY(BLIST_BORKEN) },
	{ ENTRY(BLIST_KEY) },
	{ ENTRY(BLIST_SINGLELUN) },
	{ ENTRY(BLIST_NOTQ) },
	{ ENTRY(BLIST_SPARSELUN) },
	{ ENTRY(BLIST_MAX5LUN) },
	{ ENTRY(BLIST_ISROM) },
	{ ENTRY(BLIST_LARGELUN) },
	{ ENTRY(BLIST_INQUIRY_36) },
	{ ENTRY(BLIST_INQUIRY_58) },
	{ ENTRY(BLIST_NOSTARTONADD) },
	{ ENTRY(BLIST_MS_SKIP_PAGE_08) },
	{ ENTRY(BLIST_MS_SKIP_PAGE_3F) },
	{ ENTRY(BLIST_USE_10_BYTE_MS) },
	{ ENTRY(BLIST_MS_192_BYTES_FOR_3F) },
	{ ENTRY(BLIST_REPORTLUN2) },
	{ ENTRY(BLIST_NOREPORTLUN) },
	{ ENTRY(BLIST_NOT_LOCKABLE) },
	{ ENTRY(BLIST_NO_ULD_ATTACH) },
	{ ENTRY(BLIST_SELECT_NO_ATN) },
};

#define	dprintf(format, args...)	\
	if (debug) { log_message ("%s: " format, __FUNCTION__, args); }

static void log_message (const char *format, ...)
{
	va_list args;
	va_start (args, format);
	vfprintf(stderr, format, args);
}

static int flags_to_value(char *flags, int *value)
{
	char *cur, *end;
	unsigned int ind;

	/*
	 * Strip leading and trailing spaces.
	 */
	cur = flags;
	while (*cur != '\0' && isspace(*cur))
		cur++;
	end = cur + strlen(cur) - 1;
	while (isspace(*end))
		*end-- = '\0';

	*value = 0;
	for (ind = 0;
		ind < sizeof(devinfo_to_value)/sizeof(devinfo_to_value[0]);
		ind++) {
			if (!(strcmp(devinfo_to_value[ind].name, cur))) {
				*value = devinfo_to_value[ind].value;
				return 0;
			}
	}

	log_message("File %s line %d: Could not match flag '%s'\n",
		     cur_filename, cur_lineno, cur);
	return 1;
}

static int get_flags(char *string, int *bflags)
{
	char *cur;
	int flag;

	*bflags = 0;
	cur = string;
	while (cur && cur[0] != '\0') {
		cur = strsep(&string, "|");
		if (cur) {
			if (flags_to_value(cur, &flag))
				return 1;
			*bflags = *bflags | flag;
		}
	}
	return 0;
}

static int get_token (char **token, char **line, unsigned int max,
		      char *token_name)
{
	char *cur_token;

	cur_token = strsep(line, ":");
	if (!cur_token || (*cur_token == '\0')) {
		log_message("File %s line %d: '%s' token bad or missing\n",
			     cur_filename, cur_lineno, token_name);
		return 1;
	}
	*token = cur_token;
	if (max && (strlen(*token) > max)) {
		log_message("File %s line %d: '%s' value '%s'"
			    " is longer than max %d\n", cur_filename,
			    cur_lineno, token_name, *token, max);
		return 1;
	}
	return 0;
}

static int parse_and_dump(char *line, char *result)
{
	char *vendor, *model, *flags;
	int bflags;

	dprintf("line %d: '%s'\n", cur_lineno, line);

#define get_and_test_token(token, max) \
	if (get_token(&token, &line, max, #token)) { \
		return 1; \
	} \

	get_and_test_token(vendor, 8);
	get_and_test_token(model, 16);
	get_and_test_token(flags, 0);

#undef get_and_test_token

	/*
	 * Error out if anything is left.
	 */
	if (line) {
		log_message("File %s line %d: extra tokens or data\n",
			     cur_filename, cur_lineno);
		return 1;
	}

	if (get_flags(flags, &bflags))
		return 1;

	sprintf(result, "%s:%s:0x%x", vendor, model, bflags);
	return 0;
}

static size_t buffer_getline(char **line, size_t *len, char **file_buffer,
			     char *endbuf)
{
	char *buf, *curline, *endline, *prev_line;
	size_t res;

	/*
	 * Sort of a getline for use with mmap, see man getline. Idea and
	 * some code courtesy of Kay Sievers + udev code: since klibc
	 * doesn't have getline and its fgets does/did IO one byte at a
	 * time, use mmap.
	 */

	if (!*line)
		/* 
		 * Firsttime. Allocation is done below only if *len is set
		 * to zero.
		 */
		*len = 0; 
	buf = *file_buffer;
	curline = *line;
	endline = *line + *len;
	while (buf < endbuf) {
		if (curline >= endline) {
			*len += GETLINE_CHUNK_SIZE;
			prev_line = *line;
			*line = realloc(*line, *len);
			if (!*line) {
				log_message("Unable to allocate %d bytes\n",
					     *len);
				exit(1);
			}
			curline = *line + (curline - prev_line);
			endline = *line + *len;
		}
		*curline = *buf;
		curline++;
		if (*buf++ == '\n')
			break;
	}

	res = buf - *file_buffer;
	*file_buffer = buf;
	if (res == 0)
		res = -1;
	/* 
	 * dprintf("returning res %d; *line 0x%x; *len %d\n", res, *line, *len);
	 */
	return res;
}

static void read_and_write(char *file_buffer, size_t file_size)
{
	char *line, *cur, *result, *end_buffer;
	size_t len, max_len;
	int res;
	int out_fd;

	cur_lineno = 0; /* global */
	max_len = 0;
	result = NULL;
	line = NULL;
	end_buffer = file_buffer + file_size;
	/*
	 * buffer_getline modifies file_buffer, so end_buffer is needed.
	 */
	while ((res = buffer_getline(&line, &len, &file_buffer,
				     end_buffer)) != -1) {
		if (line[res - 1] == '\n')
			line[res - 1] = '\0';
		cur = line;
		cur_lineno++;
		while (*cur != '\0' && isspace(*cur))
			cur++;
		if ((*cur == '\0') || (*cur == '#'))
			continue;

		if (len > max_len)  {
			/*
			 * result is NULL the first time, realloc() is OK
			 * with that.
			 */
			result = realloc(result, len);
		}
		if (!result) {
			log_message("Unable to allocate %d bytes\n", len);
			exit(1);
		}
		max_len = len;
		if (parse_and_dump(line, result)) {
			log_message("Parse failure: %s at line number %d\n",
				     cur_filename, cur_lineno);
			continue;
		}

		if (test_only) {
			log_message("write '%s' to '%s'\n", result, out_file);
			continue;
		}

		dprintf("Writing '%s' to '%s'\n", result, out_file);
		out_fd = open(out_file, O_WRONLY);
		if (out_fd == -1) {
			log_message("Cannot open '%s', error: %s.\n", out_file,
				    strerror(errno));
			exit(1);
		}
		if (write(out_fd, result, strlen(result)) == -1) {
			log_message("write failed to '%s', error: %s.\n",
				    out_file, strerror(errno));
			exit(1);
		}
		if (close(out_fd) == -1) {
			log_message("Cannot close '%s', error: %s.\n",
				    out_file, strerror(errno));
			exit(1);
		}
	}
	free(line);
	free(result);
}

/*
 * Open cur_filename.
 */
static void open_and_parse(void)
{
	char *file_buffer;
	int fd;
	size_t file_size;
	struct stat stats;

	dprintf("%s\n", cur_filename);

	fd = open(cur_filename, O_RDONLY);
	if (fd < 0) {
		log_message("Cannot open '%s', error: %s.\n", cur_filename,
			    strerror(errno));
		exit(1);
	}
	if (fstat(fd, &stats) < 0) {
		log_message("Cannot fstat '%s', error: %s.\n", cur_filename,
			    strerror(errno));
		exit(1);
	}
	file_buffer = mmap(NULL, stats.st_size, PROT_READ, MAP_SHARED, fd, 0);
	if (file_buffer == MAP_FAILED) {
		log_message("Cannot mmap '%s', error: %s.\n", cur_filename,
			    strerror(errno));
		exit(1);
	}
	file_size = stats.st_size;

	read_and_write(file_buffer, file_size);

	close(fd);
}

static void find_files(void)
{
	DIR *dir;
	struct dirent *ent;
	char *file_suffix;
	int name_len;

	dprintf("scsidir %s\n", scsidir);
	dir = opendir(scsidir);
	if (dir == NULL) {
		log_message("Cannot open '%s', error: %s.\n", scsidir,
			    strerror(errno));
		exit(1);
	}

	name_len = 0;
	while (1) {
		ent = readdir(dir);
		if ((ent == NULL) || (ent->d_name[0] == '\0'))
			break;

		if (ent->d_name[0] == '.')
			continue;

		file_suffix = strrchr(ent->d_name, '.');
		if (file_suffix == NULL)
			continue;

		if (strcmp(file_suffix, DEVINFO_SUFFIX))
			continue;

		snprintf(cur_filename, MAX_NAME_LEN, "%s/%s", scsidir,
			 ent->d_name);
		open_and_parse();
	}

	closedir(dir);
}

int main (int argc, char **argv)
{
	int option;

	while (1) {
		option = getopt(argc, argv, options);
		if (option == -1)
			break;

		switch (option) {
		case 'd':
			strncpy(scsidir, optarg, MAX_NAME_LEN);
			break;

		case 'p':
			strncpy(out_file, optarg, MAX_NAME_LEN);
			break;

		case 't':
			test_only = 1;
			break;

		case 'V':
			debug = 1;
			break;

		default:
			fprintf(stderr, "Invalid option '%c'\n", option);
			exit(1);
		}
	}

	find_files();
	exit (0);
}
